package com.gxzc.zen.common.config

import com.github.benmanes.caffeine.cache.Caffeine
import com.gxzc.zen.common.persistance.CacheSpec
import org.springframework.boot.autoconfigure.condition.ConditionalOnProperty
import org.springframework.boot.context.properties.ConfigurationProperties
import org.springframework.cache.CacheManager
import org.springframework.cache.annotation.EnableCaching
import org.springframework.cache.caffeine.CaffeineCache
import org.springframework.cache.support.SimpleCacheManager
import org.springframework.context.annotation.Bean
import org.springframework.context.annotation.Configuration
import org.springframework.context.annotation.Primary
import java.util.concurrent.TimeUnit

/**
 * 缓存配置类
 * @author NorthLan at 2018/2/4
 */
@Configuration
@EnableCaching
@ConfigurationProperties(prefix = "cache")
@ConditionalOnProperty(prefix = "cache", name = ["enable"], havingValue = "true")
class CacheConfig {

    var cacheSpecs: Map<String, CacheSpec>? = null

    @Bean
    @Primary
    fun caffeineCacheManager(): CacheManager {
        return SimpleCacheManager().also {
            it.setCaches(buildCache())
        }
    }

    fun buildCache(): MutableCollection<CaffeineCache> {
        val collection = mutableListOf<CaffeineCache>()
        cacheSpecs?.forEach { name, spec ->
            run {
                val builder = Caffeine.newBuilder().also {
                    if (spec.initialCapacity >= 0) {
                        it.initialCapacity(spec.initialCapacity)
                    }
                    if (spec.maximumSize >= 0) {
                        it.maximumSize(spec.maximumSize)
                    }
                    if (spec.maximumWeight >= 0) {
                        it.maximumWeight(spec.maximumWeight)
                    }
                    if (spec.expireAfterAccess >= 0) {
                        it.expireAfterAccess(spec.expireAfterAccess, TimeUnit.SECONDS)
                    }
                    if (spec.expireAfterWrite >= 0) {
                        it.expireAfterWrite(spec.expireAfterWrite, TimeUnit.SECONDS)
                    }
                    if (spec.refreshAfterWrite >= 0) {
                        it.refreshAfterWrite(spec.refreshAfterWrite, TimeUnit.SECONDS)
                    }
                    if (spec.weakKeys) {
                        it.weakKeys()
                    }
                    if (spec.weakValues) {
                        it.weakValues()
                    }
                    if (spec.softValues) {
                        it.softValues()
                    }
                    if (spec.recordStats) {
                        it.recordStats()
                    }
                }
                collection.add(CaffeineCache(name, builder.build()))
            }
        }
        return collection
    }
}